unit uMy2DMatrix;

interface
uses
  System.Sysutils;

const
  C_DefaultCellFormat = '%5.5g';
type
  EMy2dMatrixException = class(Exception);
  TDataArray = array of array of Extended;

  TMy2DMatrix = record
    class operator Multiply(a: Extended; b: TMy2DMatrix): TMy2DMatrix;
    class operator Multiply(b: TMy2DMatrix; a: Extended): TMy2DMatrix;
    class operator Multiply(a: TMy2DMatrix; b: TMy2DMatrix) : TMy2DMatrix;
    class operator Add(a: TMy2DMatrix; b: TMy2DMatrix) : TMy2DMatrix;
    class operator Subtract(a: TMy2DMatrix; b: TMy2DMatrix) : TMy2DMatrix;
    class operator Implicit(a: TMy2DMatrix): string;
    class operator Equal(a, b: TMy2DMatrix): boolean;
  private
    FRows: Integer;
    FColumns: Integer;
    FData: TDataArray;
    FCellFormat: string;
    function TestCellFormat(const AFmt: string): Boolean;
    function GetData(const Row, Col: Integer): Extended;
    procedure SetData(const Row, Col: Integer; const Value: Extended);
    function GetTranspose: TMy2DMatrix;
  public
    function AsText(Precision: string = C_DefaultCellFormat): string;
    procedure Initialize(const ARows, ACols: Integer; const ACellFormat: string = C_DefaultCellFormat);
    property Transpose: TMy2DMatrix read GetTranspose;
    property Rows: Integer read FRows;
    property Columns: Integer read FColumns;
    property Data[const Row, Col: Integer]: Extended read GetData write SetData; default;
    property CellFormat: string read FCellFormat write FCellFormat;
  end;

  TMatrixTest = class
  private
    FMatrix: TMy2DMatrix;
    function GetColumns: Integer;
    function GetRows: Integer;
    function GetTranspose: TMy2DMatrix;
    function GetCellFormat: string;
    function GetData(const Row, Col: Integer): Extended;
    procedure SetCellFormat(const Value: string);
    procedure SetData(const Row, Col: Integer; const Value: Extended);
  public
    property Data[const Row, Col: Integer]: Extended read GetData write SetData;
    function AsText(Precision: string = C_DefaultCellFormat): string;
    procedure Initialize(const ARows, ACols: Integer; const ACellFormat: string = C_DefaultCellFormat);
    property Transpose: TMy2DMatrix read GetTranspose;
    property Rows: Integer read GetRows;
    property Columns: Integer read GetColumns;
    property CellFormat: string read GetCellFormat write SetCellFormat;
    property Matrix: TMy2DMatrix read FMatrix;
  end;

implementation

const
  crlf = #13#10;

{ TMatrix }

class operator TMy2DMatrix.Add(a, b: TMy2DMatrix): TMy2DMatrix;
var
  I: Integer;
  J: Integer;
begin
  if (a.Rows = b.Rows) and (a.Columns = b.Columns) then
  begin
    Result.Initialize(a.Rows, a.Columns);
    for I := 0 to a.Rows-1 do
      for J := 0 to a.Columns-1 do
        Result.FData[I, J] := a.FData[I, J] + b.FData[I, J];
  end
  else
    raise EMy2dMatrixException.Create(
      'Matrix A and B must both have the same dimensions to add them together.');
end;

function TMy2DMatrix.AsText(Precision: string = '%5.5g'): string;
var
  J, I, LLen: Integer;
  LLEnd: string;
  LREnd: string;
  LTop: string;
begin
  if TestCellFormat(Precision) then
  begin
    // method to report a matrix as a formatted multi-line string.
    if (Rows = 0) or (Columns = 0) then
      Result := Concat('Null matrix', crlf)
    else
    begin
      Result := '';
      LLen := 0;
      for J := 0 to Rows-1 do
      begin
        if J < (Rows-1) then
        begin
          LLEnd := '| ';
          LREnd := ' |';
        end
        else
        begin
          LLEnd := '|_';
          LREnd := '_|';
        end;

        Result := Concat(Result, LLEnd, Format(Precision, [FData[J, 0]]));
        for I := 1 to Columns-1 do
          Result := Concat(Result, ', ', Format(Precision, [Data[J, I]]));
        Result := Concat(Result, LREnd, crlf);
        if LLen = 0 then
        begin
          LLen := Length(Result)-6;
          LTop := format(' _%*.*s_%s', [LLen, LLen, ' ', crlf]);
        end;
      end;

      Result := Concat(LTop, Result, crlf);
    end;
  end;
end;

class operator TMy2DMatrix.Equal(a, b: TMy2DMatrix): boolean;
var
  I: Integer;
  J: Integer;
begin
  // for two matrices to be "Equal", they must both have the same number
  // of rows and columns and each cell in one matrix must have the same value
  // as the corresponding cell in the other matrix.
  Result := (a.Rows = b.Rows) and (a.Columns = b.Columns);
  if Result then
  begin
    for I := 0 to a.Rows-1 do
      for J := 0 to a.Columns-1 do
      begin
        Result := a[I, J] = b[I, J];
        if not Result then
          Exit;
      end;
  end;
end;

function TMy2DMatrix.GetData(const Row, Col: Integer): Extended;
begin
  Result := FData[Row, Col];
end;

function TMy2DMatrix.GetTranspose: TMy2DMatrix;
var
  I: Integer;
  J: Integer;
begin
  Result.Initialize(Columns, Rows);
  for I := 0 to Columns-1 do
    for J := 0 to Rows-1 do
      Result[I, J] := Data[J, I];
end;

class operator TMy2DMatrix.Implicit(a: TMy2DMatrix): string;
begin
  Result := a.AsText(a.CellFormat);
end;

procedure TMy2DMatrix.Initialize(const ARows, ACols: Integer; const ACellFormat: string);
begin
  SetLength(FData, ARows, ACols);
  FRows := ARows;
  FColumns := ACols;
  FCellFormat := ACellFormat;
end;

class operator TMy2DMatrix.Multiply(a: Extended; b: TMy2DMatrix): TMy2DMatrix;
var
  I: Integer;
  J: Integer;
begin
  // Scalar multiplication aB. This involves multiplying each element of the
  // matrix represented by B by the value represented by A.
  Result.Initialize(b.Rows, b.Columns);
  for I := 0 to b.Rows-1 do
    for J := 0 to b.Columns-1 do
      Result.FData[I, J] := b.FData[I, J] * a;
end;

class operator TMy2DMatrix.Multiply(a, b: TMy2DMatrix): TMy2DMatrix;
var
  I: Integer;
  J: Integer;
  K: Integer;
  LValue: Extended;
begin
  //  If A is an nm matrix and B is an mp matrix, the result AB of
  //  their multiplication is an np matrix, defined only if the number
  //  of columns m in A is equal to the number of rows m in B.
  if (a.Columns = b.Rows) then
  begin
    Result.Initialize(a.Rows, b.Columns);

    // I, J are indexes into the Result array
    for I := 0 to a.Rows-1 do
      for J := 0 to b.Columns-1 do
      begin
        LValue := 0;
        // K iterates over the columns in A and rows in B.
        for K := 0 to a.Columns-1 do
          LValue := LValue + a.FData[I,K] * b.FData[K,J];
        Result.FData[I, J] := LValue;
      end;
  end
  else
    raise EMy2dMatrixException.CreateFmt(
      'Matrix A has dimensions (%d x %d). Matrix B is (%d x %d). A must '+
      'have the same number of rows as B has columns',
      [a.Rows, a.Columns, b.Rows, b.Columns]);
end;

class operator TMy2DMatrix.Multiply(b: TMy2DMatrix; a: Extended): TMy2DMatrix;
begin
  // scalar multiplication of a matrix is commutative. i.e aB = Ba. This
  // method provides the rules for the compiler to understand that.
  Result := a * b;
end;

procedure TMy2DMatrix.SetData(const Row, Col: Integer; const Value: Extended);
begin
  FData[Row, Col] := Value;
end;

class operator TMy2DMatrix.Subtract(a, b: TMy2DMatrix): TMy2DMatrix;
var
  I: Integer;
  J: Integer;
begin
  if (a.Rows = b.Rows) and (a.Columns = b.Columns) then
  begin
    Result.Initialize(a.Rows, a.Columns);
    for I := 0 to a.Rows-1 do
      for J := 0 to a.Columns-1 do
        Result.FData[I, J] := a.FData[I, J] - b.FData[I, J];
  end
  else
    raise EMy2dMatrixException.Create(
      'Matrix A and B must both have the same dimensions to subtract one from the other.');
end;

function TMy2DMatrix.TestCellFormat(const AFmt: string): Boolean;
var
  LTestStr: string;
begin
  try
    LTestStr := format(AFmt, [Pi]);
    Result := Pos('3.1', LTestStr) > 0;
  except
    on Exception do
      Result := False;
  end;

end;

{ TMatrixTest }

function TMatrixTest.AsText(Precision: string): string;
begin
  Result := FMatrix.AsText(Precision);
end;

function TMatrixTest.GetCellFormat: string;
begin
  Result := FMatrix.CellFormat;
end;

function TMatrixTest.GetColumns: Integer;
begin
  Result := FMatrix.Columns;
end;

function TMatrixTest.GetData(const Row, Col: Integer): Extended;
begin
  Result := FMatrix[Row, Col];
end;

function TMatrixTest.GetRows: Integer;
begin
  Result := FMatrix.Rows;
end;

function TMatrixTest.GetTranspose: TMy2DMatrix;
begin
  Result := FMatrix.Transpose;
end;

procedure TMatrixTest.Initialize(const ARows, ACols: Integer;
  const ACellFormat: string);
begin
  FMatrix.Initialize(ARows, ACols, ACellFormat);
end;

procedure TMatrixTest.SetCellFormat(const Value: string);
begin
  FMatrix.CellFormat := Value;
end;

procedure TMatrixTest.SetData(const Row, Col: Integer; const Value: Extended);
begin
  FMatrix.Data[Row, Col] := Value;
end;

end.
